"""A PushTypeLibrary describes the PushTypes which a given instance of PushGP will support."""
from typing import Any, Sequence, Tuple, Optional, Set, Callable

from pyshgp.push.types import PushType, CORE_PUSH_TYPES
from pyshgp.push.atoms import Atom, Literal
from pyshgp.validation import PushError


RESERVED_PSEUDO_STACKS = {"exec", "untyped", "stdout"}


class PushTypeLibrary(dict):
    """A collection of PushTypes which can support a corresponding PushStack.

    Parameters
    ----------
    register_core : bool, optional
        If True, all core types will be registered. Default is True.
    *args
        A collection of PushTypes to register.

    Attributes
    ----------
    register_core : bool, optional
        If True, all core types will be registered. Default is True.
    *args
        A collection of PushTypes to register.

    """

    def __init__(self, register_core: bool = True, *args):
        if register_core:
            self.register_core()
        self.register_list(args)
        self.create_and_register("code", (Atom, ), force=True)
        self.create_and_register("exec", (Atom, ), force=True)

    def register(self, push_type: PushType, force=False):
        """Register a PushType object.

        Parameters
        ----------
        push_type
            PushType to register.

        Returns
        -------
        PushTypeLibrary
            A reference to the PushTypeLibrary.

        """
        name = push_type.name
        if (not force) and (name in RESERVED_PSEUDO_STACKS):
            raise ValueError("Cannot register PushType with name {nm} because it is reserved.".format(nm=name))
        self[name] = push_type
        return self

    def create_and_register(
        self,
        name: str,
        underlying_types: Tuple[type],
        coercion_func: Optional[Callable[[Any], Any]] = None,
        force=False
    ):
        """Create a PushType and register it into the library.

        Parameters
        ----------
        name : str
            A name for the type. Used when referencing the PushType in Instruction
            definitions and will be the key in the PushState for the corresponding
            PushStack.
        underlying : Tuple[type]
            A tuple of python types that correspond to the underlying
            native types which the PushType is representing.
        coercion_func : Callable[[Any], Any], optional
            A function which takes a single argument and returns argument coerced
            into the PushTypes canonical type (the first type in ``underlying``).
            If None, the constructor of the canonical type is used. Default is None.
        force : bool
            If True, will register the type even if it will overwrite an
            existing reserved stack typed (eg. exec, stdout, untyped). Default
            is False. It is not reccomended this argument be changed unless
            you have a very good reason to do so.

        Returns
        -------
        PushTypeLibrary
            A reference to the PushTypeLibrary.

        """
        self.register(PushType(name, underlying_types, coercion_func), force)
        return self

    def unregister(self, push_type_name: str):
        """Unregister a push type by name.

        Parameters
        ----------
        push_type_name
            The name of the push type to unregister.

        Returns
        -------
        PushTypeLibrary
            A reference to the PushTypeLibrary.

        """
        if push_type_name in RESERVED_PSEUDO_STACKS:
            raise ValueError("Cannot unregister PushType with name {nm} because it is reserved.".format(nm=push_type_name))
        self.pop(push_type_name, None)
        return self

    def register_list(self, list_of_push_types: Sequence[PushType]):
        """Register a list of PushType ojbects.

        Parameters
        ----------
        list_of_push_types
            List of Instruction objects to register.

        Returns
        -------
        PushTypeLibrary
            A reference to the PushTypeLibrary.

        """
        for push_type in list_of_push_types:
            self.register(push_type)
        return self

    def register_core(self):
        """Register all core PushTypes defined in pyshgp.

        Returns
        -------
        PushTypeLibrary
            A reference to the PushTypeLibrary.

        """
        for push_type in CORE_PUSH_TYPES:
            self.register(push_type)

    def supported_stacks(self) -> Set[str]:
        """All stack names which the PushTypeLibrary can support.

        Returns
        -------
        set[str]
            A set of stacks names which the type library can support.

        """
        return set(self.keys())

    def push_type_of(self, thing: Any, error_on_not_found: bool = False) -> Optional[PushType]:
        """Return the PushType of the given thing.

        Parameters
        ----------
        thing : Any
            Any value to try and find the corresponding PushType.
        error_on_not_found : bool, optional
            If True, will raise error if no PushType found. Default is False.

        Returns
        -------
        Optional[PushType]
            The corresponding PushType of the thing. If no corresponding type, returns None.

        """
        return self.push_type_for_type(type(thing), error_on_not_found)

    def push_type_for_type(self, typ: type, error_on_not_found: bool = False) -> Optional[PushType]:
        """Return the PushType of the given python (or numpy) type.

        Parameters
        ----------
        typ : type
            Any type to try and find the corresponding PushType.
        error_on_not_found : bool, optional
            If True, will raise error if no PushType found. Default is False.

        Returns
        -------
        Optional[PushType]
            The corresponding PushType of the given type. If no corresponding type, returns None.

        """
        for push_type in self.values():
            if typ in push_type.underlying:
                return push_type
        if error_on_not_found:
            raise PushError.no_type(typ)


def infer_literal(val: Any, type_library: PushTypeLibrary) -> Literal:
    """Make a literal by infering the PushType of the value.

    Parameters
    ----------
    val : Any
        Any value to try and make a Literal out of.
    type_library : PushTypeLibrary
        The library of PushTypes which a Literal can be made of.

    Returns
    -------
    Literal
        The Literal object which holds the value and the corresponding PushType.

    """
    return Literal(val, type_library.push_type_of(val, error_on_not_found=True))



"""Definitions for all core I/O instructions, including input instructions."""
from typing import Sequence, Callable

from pyshgp.push.state import PushState
from pyshgp.push.atoms import Atom, Literal
from pyshgp.push.instruction import SimpleInstruction, TakesStateInstruction
from pyshgp.push.type_library import PushTypeLibrary


def _nth_inputer(ndx: int) -> Callable:
    # @TODO: Replace with partial
    def f(state: PushState) -> Sequence[Literal]:
        input_value = state.inputs[ndx]
        if isinstance(input_value, Atom):
            return input_value,
        return input_value,
    return f


def make_input_instruction(ndx: int) -> TakesStateInstruction:
    """Return insctuction to push a copy of the input value at the given index."""
    return TakesStateInstruction(
        "input_{i}".format(i=ndx),
        _nth_inputer(ndx),
        output_stacks=["untyped"],
        other_stacks=[],
        code_blocks=0,
        docstring="Push a copy of input at index {i}.".format(i=ndx)
    )


def make_input_instructions(num_inputs: int) -> Sequence[TakesStateInstruction]:
    """Return insctuctions to push a copy of the input value at the given index."""
    return [make_input_instruction(i) for i in range(num_inputs)]


# Printing instructions

def instructions(type_library: PushTypeLibrary):
    """Return all core printing instructions."""
    i = []

    for push_type in type_library.keys():
        i.append(SimpleInstruction(
            "print_{t}".format(t=push_type),
            lambda x: [str(x)],
            input_stacks=[push_type],
            output_stacks=["stdout"],
            code_blocks=0,
            docstring="Prints the top {t}.".format(t=push_type)
        ))
    return i



"""pyshgp.push.instructions."""
from typing import Sequence, Set
from itertools import chain

from pyshgp.push.type_library import PushTypeLibrary
from pyshgp.push.atoms import Instruction
from pyshgp.push.instructions import common, numeric, text, code, io, logical


def _supported(instructions: Sequence[Instruction], supported_stacks: Set[str]) -> Sequence[Instruction]:
    supported = []
    for instr in instructions:
        if instr.required_stacks() <= supported_stacks:
            supported.append(instr)
    return supported


def generic_instructions(type_library: PushTypeLibrary) -> Sequence[Instruction]:
    """All instructions that can be applied to any stack, regardless of PushType."""
    return list(chain(
        io.instructions(type_library),
        common.instructions(type_library),
    ))


def core_instructions(type_library: PushTypeLibrary) -> Sequence[Instruction]:
    """All instructions definied by pyshgp for the given type library."""
    supported_stacks = type_library.supported_stacks()
    instrucion_modules = [numeric, text, logical, code]
    basic_instr = list(chain(*[m.instructions() for m in instrucion_modules]))
    gen_instrs = generic_instructions(type_library)
    supported_instructions = _supported(basic_instr + gen_instrs, supported_stacks)
    return supported_instructions



# -*- coding: utf-8 -*-
"""The :mod:`stack` module defines the ``PushStack`` class.

A PushStack is used to hold values of a certain PushType in a PushState object.
"""
from typing import Optional, Sequence

from pyshgp.push.types import PushType
from pyshgp.utils import Token


class PushStack(list):
    """Stack that holds elements of a sinlge PushType.

    Parameters
    ----------
    push_type : PushType
        The PushType all items of the stack should conform to.
    values: Sequence, optional
        A collection of values to push onto the stack. Values will be pushed in
        order resulting in the first item being at the bottom of the stack and the
        last item being on top.

    Attributes
    ----------
    push_type : PushType
        The PushType all items of the stack should conform to.

    """

    __slots__ = ["push_type"]

    def __init__(self, push_type: PushType, values: Optional[Sequence] = None):
        self.push_type = push_type
        if values is not None:
            for val in values:
                self.push(val)

    def is_empty(self) -> bool:
        """Return True if the stack is empty. Return False otherwise."""
        return len(self) == 0

    def _coerce(self, value):
        if not self.push_type.is_instance(value):
            value = self.push_type.coerce(value)
        return value

    def push(self, value):
        """Pushes a value to the top of the stack.

        A type safe alias for `append` which makes the definition of push
        instructions more clear.

        Parameters
        ----------
        value :
            Value to push onto stack.

        """
        self.append(self._coerce(value))
        return self

    def pop(self, index: Optional[int] = None):
        """Pop the top item off the stack, or pop the item at some index.

        Parameters
        ----------
        index :
            Index to pop from the stack.

        Returns
        --------
        Element at ``index`` in stack.

        """
        if index is None:
            return super().pop()
        else:
            return super().pop((len(self) - 1) - index)

    def nth(self, position: int):
        """Return the element at a given position.

        If stack is empty, returns None. If ``position < 0`` or
        ``position > len(self)`` returns a ``no_stack_item`` token.

        Parameters
        ----------
        position : int
            Position in stack to get item.

        Returns
        --------
        Element at ``positon`` in stack.

        """
        if len(self) <= position:
            return Token.no_stack_item
        elif len(self) == 0:
            return Token.no_stack_item
        else:
            return self[(len(self) - 1) - position]

    def take(self, n: int):
        """Return the top `n` items from the tack.

        If `n` is less than zero, result is an empty list. If `n` is greater
        than the size of the stack, result contains every element.

        Parameters
        ----------
        n : int
            Number of items

        Returns
        --------
        Element at ``positon`` in stack.

        """
        if n < 0:
            return []
        return self[:-(n + 1):-1]

    def top(self):
        """Return the top item on the stack, or a ``no_stack_item`` token if empty.

        Returns
        --------
        Returns the top element of the stack, or None if empty.

        """
        if len(self) > 0:
            return self[-1]
        else:
            return Token.no_stack_item

    def insert(self, position: int, value):
        """Insert value at position in stack.

        Parameters
        ----------
        position : int
            Position in stack to get item.
        value :
            Value to insert into stack.

        """
        value = self._coerce(value)
        super().insert(len(self) - position, value)
        return self

    def set_nth(self, position: int, value):
        """Overwrite the item in the nth position of the stack with the new value.

        Parameters
        ----------
        position : int
            Position in stack to get item.
        value :
            Value to insert into stack.

        """
        value = self._coerce(value)
        self[len(self) - 1 - position] = value
        return self

    def flush(self):
        """Empty the stack."""
        del self[:]
        return self

    def __repr__(self):
        self_r = list(self)[::-1]
        return self_r.__repr__()

    def __eq__(self, other):
        if not isinstance(other, PushStack):
            return False
        return self.push_type == other.push_type and list(self) == list(other)



"""The ``types`` module contains the core PushTypes and functions to reference them.

A PushType simply is a named collection of Python types. A PushType
can be used to determine if multiple items should be considered the same type
during Push program execution.

"""
from typing import Any, Tuple, Callable, Optional

import numpy as np

from pyshgp.validation import PushError


class PushType:
    """Type information for values used by Push programs.

    Parameters
    ----------
    name : str
        A name for the type. Used when referencing the PushType in Instruction
        definitions and will be the key in the PushState for the corresponding
        PushStack.
    underlying : Tuple[type]
        A tuple of python (or numpy) types that correspond to the underlying
        native types which the PushType is representing.
    coercion_func : Callable[[Any], Any], optional
        A function which takes a single argument and returns argument coerced
        into the PushTypes canonical type (the first type in ``underlying``).
        If None, the constructor of the canonical type is used. Default is None.

    Attributes
    ----------
    name : str
        A name for the type. Used when referencing the PushType in Instruction
        definitions and will be the key in the PushState for the corresponding
        PushStack.
    underlying : Tuple[type]
        A tuple of python (or numpy) types that correspond to the underlying
        native types which the PushType is representing.
    coercion_func : Callable[[Any], Any], optional
        A function which takes a single argument and returns argument coerced
        into the PushTypes canonical type (the first type in ``underlying``).
        If None, the constructor of the canonical type is used. Default is None.

    """

    def __init__(self, name: str, underlying: Tuple[type], coercion_func: Optional[Callable[[Any], Any]] = None):
        self.name = name
        self.underlying = underlying
        if coercion_func is None:
            self.coercion_func = underlying[0]
        else:
            self.coercion_func = coercion_func

    def is_instance(self, thing: Any) -> bool:
        """Return true if thing is instance of PushTypes underlying type(s)."""
        return isinstance(thing, self.underlying)

    def coerce(self, thing: Any):
        """Convert thing into PushTypes underlying type."""
        try:
            return self.coercion_func(thing)
        except Exception as e:
            err_type = type(e).__name__
            err_msg = str(e)
            raise PushError(
                "{t} while coerceing {x} of {typ1} to {typ2}. Origional mesage: \"{m}\"".format(
                    t=err_type, x=thing, typ1=type(thing), typ2=self, m=err_msg
                )
            )

    def __repr__(self):
        return self.name + "<" + ",".join([t.__name__ for t in list(self.underlying)]) + ">"

    def __eq__(self, other):
        if not isinstance(other, PushType):
            return False
        return self.name == other.name and self.underlying == other.underlying

    def __hash__(self):
        return (self.name + str(self.underlying[0])).__hash__()


class Char(str):
    """Holds a string of length 1.

    Used to distinguish between string and char literals in Push program
    interpretation.

    Attributes
    ----------
    char : str
        String of length 1.

    """

    def __init__(self, char):
        if len(char) == 0:
            raise PushError.empty_character()
        if len(char) > 1:
            raise PushError.long_character()

    def __eq__(self, other):
        return isinstance(other, Char) and str(self) == str(other)


# @TODO: Add vector type(s)


PushInt = PushType("int", (int, np.int64, np.int32, np.int16, np.int8))
PushFloat = PushType("float", (float, np.float64, np.float32, np.float16))
PushStr = PushType("str", (str, np.str_))
PushBool = PushType("bool", (bool, np.bool_))
PushChar = PushType("char", (Char, ))

CORE_PUSH_TYPES = [PushBool, PushInt, PushChar, PushFloat, PushStr]



"""The :mod:`instruction_set` module defines the InstructionSet class.

An InstructionSet is a collection of Instruction objects, stored by name. The
InstructionSet has methods to help define and register additional instructions.
"""
from typing import Sequence, Set
import re

from pyshgp.push.type_library import PushTypeLibrary, RESERVED_PSEUDO_STACKS
from pyshgp.push.atoms import Instruction
from pyshgp.push.instructions import core_instructions, io


class InstructionSet(dict):
    """A collection of Instruction objects stored by name.

    Parameters
    ----------
    type_library : PushTypeLibrary, optional
        The PushTypeLibrary which denote the PushTypes (and thus stacks)
        are supported. Default is None, which corresponds to the core set of types.
    register_all : bool, optional
        If True, all instructions in the core instruction set will be registered
        upon instanciation. Default is False.
    strip_docstrings : bool, optional
        If True, the docstring attribute of registered instructions will be
        removed to reduce memory footprint. Default is True.

    Attributes
    ----------
    type_library : PushTypeLibrary, optional
        The PushTypeLibrary which denote the PushTypes (and thus stacks)
        are supported. Default is None, which corresponds to the core set of types.
    register_all : bool, optional
        If True, all instructions in the core instruction set will be registered
        upon instanciation. Default is False.
    strip_docstrings : bool, optional
        If True, the docstring attribute of registered instructions will be
        removed to reduce memory footprint. Default is True.

    """

    def __init__(self, type_library: PushTypeLibrary = None, register_core: bool = False, strip_docstrings: bool = True):
        self.strip_docstrings = strip_docstrings

        if type_library is None:
            type_library = PushTypeLibrary()
        self.type_library = type_library

        if register_core:
            self.register_core()

    def set_type_library(self, type_library: PushTypeLibrary):
        """Set the type library attribute and return self.

        Parameters
        ----------
        type_library
            PushTypeLibrary to set.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        self.type_library = type_library
        return self

    def register(self, instr: Instruction):
        """Register an Instruction object.

        Parameters
        ----------
        instr
            Instruction to register.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        if self.strip_docstrings and hasattr(instr, "docstring"):
            del instr.docstring
        if instr.required_stacks() <= self.type_library.supported_stacks():
            self[instr.name] = instr
        return self

    def register_list(self, instrs: Sequence[Instruction]):
        """Register a list of Instruction ojbects.

        Parameters
        ----------
        instrs
            List of Instruction objects to register.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        for i in instrs:
            self.register(i)
        return self

    def register_core_by_stack(
        self,
        include_stacks: Set[str],
        *,
        exclude_stacks: Set[str] = None
    ):
        """Register all instructions that make use of the given type name.

        Parameters
        ----------
        type_names
            List of PushType names.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        for i in core_instructions(self.type_library):
            req_stacks = i.required_stacks()
            if req_stacks <= include_stacks:
                if exclude_stacks is not None and len(req_stacks & exclude_stacks) > 0:
                    break
                self.register(i)
        return self

    def register_core_by_name(self, name_pattern: str):
        """Register all instructions whose name match the given pattern.

        Parameters
        ----------
        name_pattern
            A regex string.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        re_pat = re.compile(name_pattern)
        for i in core_instructions(self.type_library):
            if re.match(re_pat, i.name) is not None:
                self.register(i)
        return self

    def register_core(self):
        """Register all core instructions defined in pyshgp.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        self.register_list(core_instructions(self.type_library))
        return self

    def register_n_inputs(self, n: int):
        """Create and register `n` input instructions.

        Parameters
        ----------
        n
            The number of input instructions to make.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        input_instructions = io.make_input_instructions(n)
        self.register_list(input_instructions)
        return self

    def unregister(self, instruction_name: str):
        """Unregister an instruction by name.

        Parameters
        ----------
        instruction_name
            The name of the instruciton to unregister.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        self.pop(instruction_name, None)
        return self

    def required_stacks(self) -> Set[str]:
        """Return all stack names used throughout the registered instructions.

        Returns
        -------
        Set[str]
            The set of stack names that are used by the registered instructions.

        """
        all_types = set()
        for instr in self.values():
            all_types.update(instr.required_stacks())
        return all_types - RESERVED_PSEUDO_STACKS



"""Verbosity configuration."""
from typing import Union, Callable

import logging

# @TODO: Add verbosity of serach config
# @TODO: Add verbosity of start time, end time, and runtime.


class VerbosityConfig:
    """Collection of indicators for what should be logged in a run."""

    def __init__(self,
                 solution_found: Union[bool, int] = False,
                 generation: Union[bool, int] = False,
                 every_n_generations: int = 1,
                 simplification: Union[bool, int] = False,
                 simplification_step: Union[bool, int] = False,
                 program_trace: Union[bool, int] = False):
        self.log_level = self._update_log_level()
        self.solution_found = solution_found
        self.generation = generation
        self.every_n_generations = every_n_generations
        self.simplification = simplification
        self.simplification_step = simplification_step
        self.program_trace = program_trace

    def _update_log_level(self):
        self.log_level = logging.getLogger().getEffectiveLevel()


DEFAULT_VERBOSITY_LEVELS = [
    VerbosityConfig(),
    VerbosityConfig(
        solution_found=logging.INFO,
        generation=logging.INFO,
        every_n_generations=5,
        simplification=logging.INFO,
        simplification_step=logging.DEBUG
    ),
    VerbosityConfig(
        solution_found=logging.INFO,
        generation=logging.INFO,
        simplification=logging.INFO,
        simplification_step=logging.INFO,
        program_trace=logging.DEBUG
    )
]


def log_function(level: int) -> Callable:
    """Return corresponding log function."""
    _lvl2func = {
        10: logging.debug,
        20: logging.info,
        30: logging.warning,
        40: logging.error,
        50: logging.critical,
    }
    return _lvl2func.get(level, None)


def log(level: int, msg, *args, **kwargs):
    """Log the given message at the given log level. See ``logger`` lib."""
    f = log_function(level)
    if f is not None:
        f(msg, *args, **kwargs)
